import inspect
import warnings
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator,
                             Callable, Generator, Iterable, Iterator, Mapping)
from functools import wraps
from typing import List, Tuple, Optional

from ..utils import exceptions as exc
from ..utils.compat import is_classvar, is_final
from ..utils.datastructures import cached_property, unprovided
from ..utils.functional import pop
from .base import BaseParser
from .field import ParserField
from .options import Options, RuntimeContext
from .rule import Rule, resolve_forward_type

LAMBDA_NAME = (lambda: None).__name__
LOCALS_NAME = "<locals>"


def _f_pass_doc():
    """"""


def _f_pass():
    pass


PASSED_CODES = (
    _f_pass.__code__.co_code,
    _f_pass_doc.__code__.co_code,
)


class FunctionParser(BaseParser):
    @classmethod
    def function_pass(cls, f):
        if not inspect.isfunction(f):
            return False
        return getattr(f, "__code__").co_code in PASSED_CODES

    @classmethod
    def validate_function(cls, f):
        return (
            isinstance(f, (staticmethod, classmethod))
            or inspect.ismethod(f)
            or inspect.isfunction(f)
        )

    @classmethod
    def apply_class(
        cls,
        target: type,
        options: Options = None,
        no_cache: bool = False,
        ignore_params: bool = False,
        ignore_result: bool = False,
        eager: bool = False,
    ):
        """
        Patch all explicit methods in class (name not beginning with "_")
        """
        for key, val in target.__dict__.items():
            if key.startswith("_"):
                continue
            if not cls.validate_function(val):
                continue
            current_parser = cls.resolve_parser(val)
            if current_parser:
                continue
            parser = cls.apply_for(
                val, no_cache=no_cache, options=options, from_class=target
            )
            func = parser.wrap(
                parse_params=not ignore_params,
                parse_result=not ignore_result,
                eager_parse=eager,
            )
            setattr(target, key, func)
        return target

    @classmethod
    def analyze_func(cls, f):
        first_reserve = None
        if isinstance(f, classmethod):
            first_reserve = True
            f = f.__func__
        elif isinstance(f, staticmethod):
            first_reserve = False
            f = f.__func__
        elif inspect.ismethod(f):
            first_reserve = False
        elif not inspect.isfunction(f):
            raise TypeError(f"Invalid function: {f}")
        return f, first_reserve

    @classmethod
    def get_return_type(cls, f):
        if isinstance(f, (staticmethod, classmethod)):
            f = f.__func__
        if not f:
            return None
        return getattr(f, "__annotations__", {}).get("return")

    @classmethod
    def infer_instancemethod(cls, func):
        return (
            hasattr(func, "__qualname__")
            and func.__qualname__.endswith("." + func.__name__)
            and not func.__qualname__.endswith(f"{LOCALS_NAME}." + func.__name__)
        )

    def __init__(self, func, options: Options = None, from_class: type = None):
        if not self.validate_function(func):
            raise TypeError(
                f"{self.__class__}: invalid function or method: {func}, must be method or function"
            )

        self.from_class = from_class
        self.instancemethod = False
        self.classmethod = isinstance(func, classmethod)
        self.staticmethod = isinstance(func, staticmethod)

        func, self.first_reserve = self.analyze_func(func)

        self.is_method = inspect.ismethod(func)
        self.is_lambda = inspect.isfunction(func) and func.__name__ == LAMBDA_NAME
        self.is_coroutine = inspect.iscoroutinefunction(func)
        self.is_generator = inspect.isgeneratorfunction(func)
        self.is_async_generator = inspect.isasyncgenfunction(func)
        self.is_asynchronous = self.is_coroutine or self.is_async_generator
        self.is_passed = self.function_pass(func)

        parameters = inspect.signature(func).parameters.items()

        if self.from_class:
            # within a class context, the instance method is easy to detect
            # exclude the staticmethod/classmethod is enough
            if (
                not self.first_reserve
                and not self.staticmethod
                and not self.classmethod
            ):
                self.first_reserve = True
                self.instancemethod = True
        else:
            if (
                not self.first_reserve
                and not self.is_method
                and not self.is_lambda
                and not self.staticmethod
            ):
                # guess instance method
                if len(parameters) >= 1:
                    (fk, first_param), *_rest = parameters
                    first_param: inspect.Parameter
                    if (
                        first_param.kind
                        in (
                            first_param.POSITIONAL_ONLY,
                            first_param.POSITIONAL_OR_KEYWORD,
                        )
                        and first_param.default is first_param.empty
                        and first_param.annotation is first_param.empty
                    ):

                        if self.infer_instancemethod(func):
                            self.first_reserve = True
                            self.instancemethod = True

        self.reserve_name = None
        if self.first_reserve:
            _r, *parameters = parameters
            self.reserve_name = _r[0]

        # annotates = {k: v.annotation for k, v in self.parameters if v.annotation is not v.empty}
        # defaults = {k: v.default for k, v in self.parameters if v.default is not v.empty}
        # if a param is not defaulted or annotated, it rule is Rule(require=True)

        self.exclude_indexes = set()
        self.parameters: Iterable[Tuple[str, inspect.Parameter]] = parameters
        self.kw_var = None  # only for function, **kwargs
        self.pos_var_index = None
        self.pos_var = None  # only for function, *args
        self.pos_key_map = {}  # reverse version of arg index
        self.arg_index = {}
        self.max_args = 0
        self.min_args = 0
        self.pos_annotation = None
        self.kw_annotation = None
        self.return_annotation = self.get_return_type(func)

        arg_names = []
        kw_names = []
        pos_only_keys = []
        common_arg_names = []

        for i, (k, v) in enumerate(self.parameters):
            v: inspect.Parameter
            arg_names.append(k)
            if v.kind == v.VAR_POSITIONAL:
                self.pos_var_index = i
                self.pos_var = k
                if v.annotation != v.empty:
                    self.pos_annotation = v.annotation
                continue
            elif v.kind == v.VAR_KEYWORD:
                self.kw_var = k
                if v.annotation != v.empty:
                    self.kw_annotation = v.annotation
                continue
            else:
                common_arg_names.append(k)
                if v.kind == v.POSITIONAL_ONLY:
                    pos_only_keys.append(k)
                    self.min_args += 1
                else:
                    kw_names.append(k)

                if v.kind != v.KEYWORD_ONLY:
                    self.max_args += 1
                    self.pos_key_map[i] = k
                    self.arg_index[k] = i

        self.arg_names = arg_names
        self.kw_names = kw_names
        self.pos_only_keys = pos_only_keys
        self.common_arg_names = common_arg_names

        if self.kw_var:
            opt_list = [options]
            if self.kw_annotation:
                opt_list.append(self.options_cls(addition=self.kw_annotation))
            elif self.kw_var:
                opt_list.append(self.options_cls(addition=True))
            options = self.options_cls.generate_from(*opt_list)

        super().__init__(func, options=options)

        self.init_kwargs = {"options": options, "from_class": from_class}
        if not self.kw_var and self.options.addition:
            raise exc.ConfigError(
                f"FunctionParser: {func}, specify addition: {options.addition} "
                f"without declaring the **kwargs variable"
            )
        if self.kw_var and not self.options.addition:
            warnings.warn(
                f"FunctionParser: {func}, specified **{self.kw_var}"
                f" but set addition=False, {self.kw_var} will always be empty"
            )

        if self.options.no_default:
            raise exc.ConfigError(
                f"FunctionParser: {func}.options cannot specify no_default=True"
            )

        if self.options.defer_default:
            raise exc.ConfigError(
                f"FunctionParser: {func}.options cannot specify defer_default=True"
            )

        if self.options.immutable:
            warnings.warn(
                f"FunctionParser: {func}, specified immutable=True in Options, which is useless"
            )
        if self.options.secret_names:
            warnings.warn(
                f"FunctionParser: {func}, specified secret_names in Options, which is useless"
            )

        self.position_type = None
        self.return_type = None

        self.generator_send_type = None
        self.generator_yield_type = None
        self.generator_return_type = None
        # Generator[Yield, Send, Return] or Iterator[Yield, Return]
        # self.async_generator_send_type = None
        # self.async_generator_yield_type = None
        # AsyncGenerator[Yield, Send] or AsyncIterator[Yield]

        if self.pos_annotation:
            self.position_type = self.parse_annotation(annotation=self.pos_annotation)

        self.generate_return_types()

    # @property
    # def do_parse_generator(self):
    #     return (self.is_generator or self.is_async_generator) and (
    #         self.generator_send_type
    #         or self.generator_yield_type
    #         or self.generator_return_type
    #     )

    def generate_return_types(self):
        # see if the return type match the function
        # focus on sync / async
        # generator / coroutine / async generator
        if not self.return_annotation:
            return

        self.return_type = self.parse_annotation(annotation=self.return_annotation)

        # https://docs.python.org/3/library/typing.html#typing.Generator
        if self.return_type and issubclass(self.return_type, Rule):
            if self.is_generator:
                if self.return_type.__origin__ in (Iterable, Iterator):
                    self.generator_yield_type = self.return_type.__args__[0]
                elif self.return_type.__origin__ == Generator:
                    (
                        self.generator_yield_type,
                        self.generator_send_type,
                        self.generator_return_type,
                    ) = self.return_type.__args__
                else:
                    warnings.warn(
                        f"Invalid return type annotation: {self.return_annotation} "
                        f"for generator function, should be Generator[...] / Iterator[...] / Iterable[...]"
                    )
            elif self.is_async_generator:
                if self.return_type.__origin__ in (AsyncIterable, AsyncIterator):
                    self.generator_yield_type = self.return_type.__args__[0]
                elif self.return_type.__origin__ == AsyncGenerator:
                    (
                        self.generator_yield_type,
                        self.generator_send_type,
                    ) = self.return_type.__args__
                else:
                    warnings.warn(
                        f"Invalid return type annotation: {self.return_annotation} "
                        f"for async generator function, should be "
                        f"AsyncGenerator[...] / AsyncIterator[...] / AsyncIterable[...]"
                    )

    @cached_property
    def positional_fields(self):
        fields = {}
        for index, key in self.pos_key_map.items():
            field = self.get_field(key)
            if not field:
                continue
            fields[index] = field
        return fields

    @cached_property
    def positional_only_fields(self) -> List[Tuple[int, ParserField]]:
        fields = []
        for i, key in enumerate(self.pos_only_keys):
            field = self.get_field(key)
            if not field:
                continue
            fields.append((i, field))
        return fields

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.obj.__qualname__}>"

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.obj.__qualname__}>"

    def get_pos_field(self, index: int) -> Optional[ParserField]:
        key = self.pos_key_map.get(index)
        if key:
            return self._get_field_from(self.fields, key)
        return None

    def generate_fields(self):
        exclude_vars = set()
        exclude_indexes = set()
        global_vars = self.globals
        fields = []

        for i, (name, param) in enumerate(self.parameters):
            if not self.validate_field_name(name):
                exclude_vars.add(name)
                if param.kind in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD):
                    exclude_indexes.add(i)
                continue
            if param.kind in (param.VAR_KEYWORD, param.VAR_POSITIONAL):
                continue

            annotation = None
            if param.annotation is not param.empty:
                if param.annotation is None:
                    annotation = type(None)
                    # annotation = None means no annotation (param.empty)
                else:
                    annotation = param.annotation
                    if is_final(annotation) or is_classvar(annotation):
                        warnings.warn(
                            f"{self.obj}: param: {repr(name)} invalid annotation: {annotation}, "
                            f"this is only for class variables, please use the type directly"
                        )
                        # args = get_args(annotation)
                        # annotation = args[0] if args else None
                        continue

            try:
                field = self.parser_field_cls.generate(
                    attname=name,
                    annotation=annotation,
                    default=param.default
                    if param.default is not param.empty
                    else unprovided,
                    global_vars=global_vars,
                    forward_refs=self.forward_refs,
                    options=self.options,
                    positional_only=param.kind == param.POSITIONAL_ONLY,
                    **self.kwargs
                )
            except Exception as e:
                raise exc.ConfigError(f'{self.name}: parse field [{repr(name)}] failed with error: {e}')

            fields.append(
                field
            )

        field_map = {}
        for field in fields:
            name = field.name

            if field.is_case_insensitive(self.options):
                name = name.lower()

            if name in field_map:
                raise exc.ConfigError(
                    f"{self.obj}: field name: {repr(name)} conflicted at "
                    f"{field}, {field_map[name]}",
                    obj=self.obj,
                    field=name,
                )
            if not self.is_passed:
                # is function is :pass, we do not check for now
                field.check_function(self.obj)  # check for function
            field_map[name] = field

        self.fields.update(field_map)
        self.exclude_vars = exclude_vars
        self.exclude_indexes = exclude_indexes

    def validate_fields(self):
        """
        all the non-keyword-only args need to check the
        non-keyword-only: required > has-default > no default (*pos)
        keyword-only: required/has-default > no default (**kwargs)
        """

        super().validate_fields()

        # validate DEFAULT order
        optional_name = None
        for i, (k, v) in enumerate(self.parameters):
            if self.first_reserve and not i:
                continue
            v: inspect.Parameter
            if v.kind == v.VAR_POSITIONAL:
                continue
            elif v.kind == v.VAR_KEYWORD:
                continue

            if v.kind == v.KEYWORD_ONLY:
                # keyword args does not need to check for default order
                continue

            field = self.get_field(k)
            if field:
                required = field.no_default
            else:
                required = v.default != v.empty

            if required:
                if i in self.exclude_indexes:
                    continue
                if optional_name:
                    msg = (
                        f"{self.obj}: non-default argument: {repr(k)} "
                        f"follows default argument: {repr(optional_name)}"
                    )
                    if v.kind == v.POSITIONAL_ONLY:
                        raise SyntaxError(msg)
                    else:
                        warnings.warn(msg)
            else:
                optional_name = k

    def resolve_forward_refs(self, local_vars=None, ignore_errors: bool = True):
        resolved = super().resolve_forward_refs(
            local_vars=local_vars, ignore_errors=ignore_errors
        )
        if resolved:
            if self.position_type:
                self.position_type, r = resolve_forward_type(self.position_type)
            if self.return_type:
                self.return_type, r = resolve_forward_type(self.return_type)

    def wrap(
        self,
        options: Options = None,
        first_reserve: bool = None,
        parse_params: bool = None,
        parse_result: bool = None,
        eager_parse: bool = False,
        ignore_methods: bool = False,
    ):

        if self.is_async_generator:
            f = self.get_async_generator(
                options=options,
                first_reserve=first_reserve,
                parse_params=parse_params,
                parse_result=parse_result,
                eager=eager_parse,
            )
        elif self.is_coroutine:
            f = self.get_async_call(
                options=options,
                first_reserve=first_reserve,
                parse_params=parse_params,
                parse_result=parse_result,
                eager=eager_parse,
            )
        elif self.is_generator:
            f = self.get_sync_generator(
                options=options,
                first_reserve=first_reserve,
                parse_params=parse_params,
                parse_result=parse_result,
                eager=eager_parse,
            )
        else:
            # if eager_parse:
            #     warnings.warn(f'{self.obj} is a sync function')

            @wraps(self.obj)
            def f(*args, **kwargs):  # noqa
                # MAKE CONTEXT AT RUNTIME !
                context = options.make_context() if options else self.make_context()
                return self.sync_call(
                    args,
                    kwargs,
                    context=context,
                    first_reserve=first_reserve,
                    parse_params=parse_params,
                    parse_result=parse_result,
                )

        f.__parser__ = self
        if not ignore_methods:
            if self.classmethod:
                return classmethod(f)
            elif self.staticmethod:
                return staticmethod(f)
        return f

    def parse_pos_type(self, index: int, value, context: RuntimeContext):
        options = context.options
        pos_type = self.position_type
        # we should just ignore the runtime addition type
        if not pos_type:
            return value

        pos_key = f"*{self.pos_var}:{index}" if self.pos_var else index
        with context.enter(pos_key) as new_context:
            try:
                value = new_context.transformer(value, pos_type)
            except Exception as e:
                error = exc.ParseError(
                    item=new_context.route, value=value, type=pos_type, origin_exc=e
                )
                if options.invalid_items == options.PRESERVE:
                    context.collect_waring(error.formatted_message)
                elif options.invalid_items == options.EXCLUDE:
                    context.collect_waring(error.formatted_message)
                    return unprovided
                else:
                    context.handle_error(error)
        return value

    def parse_addition(self, key: str, value, context: RuntimeContext):
        var_key = f"**{self.kw_var}:{key}" if self.kw_var else key
        return super().parse_addition(var_key, value=value, context=context)

    def parse_params(
        self, args: tuple, kwargs: dict, context: RuntimeContext
    ) -> Tuple[tuple, dict]:
        # def f(self, *args, arg1, arg2, **kwargs):
        # self.f(1, 2, 3)
        # def f(self, data):
        # self.f({})
        # not keyword-only argument may show up at pos args
        parsed_args = []
        parsed_keys = []

        # 1. parse giving args, including the positional args
        for i, arg in enumerate(args):
            if self.pos_var and i >= self.pos_var_index:
                # eg. f(a, b, *args): pos_var_index = 2
                arg = self.parse_pos_type(index=i, value=arg, context=context)
                if unprovided(arg):
                    continue
            else:
                field = self.positional_fields.get(i)

                if field:
                    if field.is_no_input(arg, options=context.options):
                        arg = field.get_default(options=context.options)
                    else:
                        parsed_keys.append(field.attname)
                        arg = field.parse_value(arg, context=context)
                    if unprovided(arg):
                        # on_error=excluded, or error collected
                        continue
                else:
                    if i in self.exclude_indexes:
                        # excluded var
                        # def f(a, _b, _c):
                        # we do not parse, just append this arg
                        pass
                    else:
                        # excess var ignore
                        continue

            parsed_args.append(arg)

        # 2. check if unprovided args has default give, and the unprovided required args
        for index, field in self.positional_only_fields:
            if field.attname in parsed_keys:
                continue
            if field.is_required(options=context.options):
                context.handle_error(exc.AbsenceError(item=field.attname))
                continue
            default = field.get_default(context.options)
            if not unprovided(default):
                # this position is definitely after parsed_args
                # because required args is always (we enforce check) ahead of default args
                parsed_args.append(default)
            parsed_keys.append(field.attname)  # need to append parsed as well
            # positional only field is excluded no matter the arg is provided or not

        parsed_kwargs = self.parse_data(
            kwargs, context=context, excluded_keys=parsed_keys, as_attname=True
        )
        context.raise_error()  # raise the parse error before calling the function
        return tuple(parsed_args), parsed_kwargs

    def get_params(
        self,
        args: tuple,
        kwargs: dict,
        context: RuntimeContext,
        first_reserve=None,
        parse_params: bool = None,
    ):
        _ = None
        if first_reserve is None:
            first_reserve = self.first_reserve
        if first_reserve:
            if self.instancemethod or self.classmethod:
                _self = pop(kwargs, "__self__")
                if self.classmethod:
                    if _self:
                        _ = getattr(_self, "__class__", None)
                    else:
                        _ = pop(kwargs, "__class__")
                elif _self:
                    _ = _self
            if args and not _:
                _, *args = args
        if parse_params:
            args, kwargs = self.parse_params(args, kwargs, context=context)
        if first_reserve:
            if self.from_class:
                if self.classmethod:
                    if not isinstance(_, type) or not issubclass(_, self.from_class):
                        raise exc.InvalidSubclass(type=self.from_class, value=_)
                else:
                    if not isinstance(_, self.from_class):
                        raise exc.InvalidInstance(type=self.from_class, value=_)
            args = (_, *args)
        return args, kwargs

    def parse_result(self, result, context: RuntimeContext):
        if self.return_type:
            try:
                result = context.transformer(result, self.return_type)
            except Exception as e:
                error = exc.ParseError(
                    item="<return>", value=result, type=self.return_type, origin_exc=e
                )
                context.handle_error(error, force_raise=True)
        return result

    def sync_from_generator(self, generator: Generator, context: RuntimeContext):
        i = 0
        sent = None
        while True:
            try:
                if sent is not None:
                    item = generator.send(sent)
                else:
                    item = next(generator)
            except StopIteration as err:
                result = err.value
                if result is None or not self.generator_return_type:
                    # raise the same StopIteration
                    return result
                try:
                    result = context.transformer(result, self.generator_return_type)
                except Exception as e:
                    error = exc.ParseError(
                        item=f"<generator.return>",
                        value=result,
                        type=self.generator_return_type,
                        origin_exc=e,
                    )
                    context.handle_error(error, force_raise=True)
                return result
            else:
                if inspect.isgenerator(item):
                    generator = item
                    continue
                    # maybe a tail opt generator

                if self.generator_yield_type:
                    try:
                        item = context.transformer(item, self.generator_yield_type)
                    except Exception as e:
                        error = exc.ParseError(
                            item=f"<generator.yield[{i}]>",
                            value=item,
                            type=self.generator_yield_type,
                            origin_exc=e,
                        )
                        context.handle_error(error, force_raise=True)

                sent = yield item

                if sent is not None:
                    if self.generator_send_type:
                        try:
                            sent = context.transformer(sent, self.generator_send_type)
                        except Exception as e:
                            error = exc.ParseError(
                                item=f"<generator.send[{i}]>",
                                value=sent,
                                type=self.generator_send_type,
                                origin_exc=e,
                            )
                            context.handle_error(error, force_raise=True)
                i += 1

    def get_sync_generator(
        self,
        options: Options = None,
        first_reserve: bool = None,
        parse_params: bool = None,
        parse_result: bool = None,
        eager: bool = False,
    ):
        @wraps(self.obj)
        def eager_generator(*args, **kwargs) -> Generator:
            context = (options or self.options).make_context()
            args, kwargs = self.get_params(
                args,
                kwargs,
                context=context,
                first_reserve=first_reserve,
                parse_params=parse_params,
            )
            func = self.obj
            generator: Generator = func(*args, **kwargs)
            if parse_result:
                return self.sync_from_generator(generator, context)
            return generator

        if eager:
            return eager_generator

        @wraps(self.obj)
        def sync_generator(*args, **kwargs):
            gen = eager_generator(*args, **kwargs)
            sent = None
            while True:
                try:
                    if sent is not None:
                        item = gen.send(sent)
                    else:
                        item = next(gen)
                except StopIteration as err:
                    return err.value
                else:
                    sent = yield item

        return sync_generator

    async def async_from_generator(
        self, generator: AsyncGenerator, context: RuntimeContext
    ):
        i = 0
        async for item in generator:
            if inspect.isasyncgen(item):
                generator = item
                continue

            if self.generator_yield_type:
                try:
                    item = context.transformer(item, self.generator_yield_type)
                except Exception as e:
                    error = exc.ParseError(
                        item=f"<asyncgenerator.yield[{i}]>",
                        value=item,
                        type=self.generator_yield_type,
                        origin_exc=e,
                    )
                    context.handle_error(error, force_raise=True)

            sent = yield item

            if sent is not None:
                if self.generator_send_type:
                    try:
                        sent = context.transformer(sent, self.generator_send_type)
                    except Exception as e:
                        error = exc.ParseError(
                            item=f"<asyncgenerator.send[{i}]>",
                            value=sent,
                            type=self.generator_send_type,
                            origin_exc=e,
                        )
                        context.handle_error(error, force_raise=True)
                # await generator.asend(sent)
                try:
                    await generator.asend(sent)
                except StopAsyncIteration:
                    return
            i += 1

    def get_async_generator(
        self,
        options: Options = None,
        first_reserve: bool = None,
        parse_params: bool = None,
        parse_result: bool = None,
        eager: bool = False,
    ):
        @wraps(self.obj)
        def eager_generator(*args, **kwargs) -> AsyncGenerator:
            context = (options or self.options).make_context()
            args, kwargs = self.get_params(
                args,
                kwargs,
                context=context,
                first_reserve=first_reserve,
                parse_params=parse_params,
            )
            func = self.obj
            generator: AsyncGenerator = func(*args, **kwargs)
            if parse_result:
                return self.async_from_generator(generator, context)
            return generator

        if eager:
            return eager_generator

        @wraps(self.obj)
        async def async_generator(*args, **kwargs):
            async_gen = eager_generator(*args, **kwargs)
            async for item in async_gen:
                sent = yield item
                if sent is not None:
                    try:
                        await async_gen.asend(sent)
                    except StopAsyncIteration:
                        return

        return async_generator

    def get_async_call(
        self,
        options: Options = None,
        first_reserve: bool = None,
        parse_params: bool = None,
        parse_result: bool = None,
        eager: bool = False,
    ):
        @wraps(self.obj)
        def eager_call(*args, **kwargs):
            context = (options or self.options).make_context()
            args, kwargs = self.get_params(
                args,
                kwargs,
                context=context,
                first_reserve=first_reserve,
                parse_params=parse_params,
            )
            if parse_result:
                return self.get_async_result(args, kwargs, context=context)
            return self.obj(*args, **kwargs)

        if eager:
            return eager_call

        @wraps(self.obj)
        async def async_call(*args, **kwargs):
            return await eager_call(*args, **kwargs)

        return async_call

    def sync_call(
        self,
        args: tuple,
        kwargs: dict,
        context: RuntimeContext,
        first_reserve=None,
        parse_params: bool = None,
        parse_result: bool = None,
    ):
        args, kwargs = self.get_params(
            args,
            kwargs,
            context=context,
            first_reserve=first_reserve,
            parse_params=parse_params,
        )
        func = self.obj
        result = func(*args, **kwargs)
        if parse_result:
            result = self.parse_result(result, context=context)
        return result

    async def get_async_result(
        self, args: tuple, kwargs: dict, context: RuntimeContext
    ):
        # we may not want to change the result form even if it's another coroutine
        # we leave to user to await it to avoid changing the actual logic of the function
        # while inspect.iscoroutine(result):
        #     result = await result
        return self.parse_result(await self.obj(*args, **kwargs), context=context)


def call(
    func: Callable,
    args=None,
    data=None,
    options=None,
    context: RuntimeContext = None,
    parser_cls=FunctionParser,
    ignore_params: bool = False,
    ignore_result: bool = False,
):

    parser = parser_cls.apply_for(func)  # use the __parser__ if already installed
    options = options or parser.options
    new_context: RuntimeContext = (options or parser.options).make_context(
        context=context
    )
    transformer = new_context.transformer

    args = args or ()
    data = data or {}
    if not isinstance(args, Iterable):
        # {} dict instance is an instance of Mapping too
        if transformer.no_explicit_cast:
            raise TypeError(
                f"invalid input type for funtional args, should be dict or Mapping"
            )
        else:
            args = transformer.to_array_types(args)
    if not isinstance(data, Mapping):
        # {} dict instance is an instance of Mapping too
        if transformer.no_explicit_cast:
            raise TypeError(
                f"invalid input type for functional data, should be dict or Mapping"
            )
        else:
            data = transformer.to_dict(data)

    if new_context.options.cast_keyword_str:
        _data = {}
        for key, val in data.items():
            if not isinstance(key, str):
                key = transformer.to_str(key)
            _data[key] = val
        data = _data

    f = parser.wrap(
        options, parse_params=not ignore_params, parse_result=not ignore_result
    )
    return f(*args, **data)



import inspect
from typing import Callable, Optional


class TypeRegistry:
    def __init__(self,
                 name: str = 'default',
                 base: 'TypeRegistry' = None,
                 validator: Callable = callable,
                 cache: bool = False,
                 default=None,
                 shortcut: str = None
                 ):
        self._registry = []
        self._cache = {}

        self.name = name
        self.cache = cache
        self.default = default
        self.shortcut = shortcut
        self.base = base
        self.validator = validator

    def register(
        self,
        *classes,
        attr=None,
        detector=None,
        metaclass=None,
        # to=None,  # register to a specific type transformer class
        allow_subclasses: bool = True,
        priority: int = 0,
    ):
        # detect class by issubclass or hasattr
        # this method can override
        # the latest function will have the final effect
        # signature = (*classes, attr, detector)

        if not detector:
            if not classes and not attr and not metaclass:
                raise ValueError(
                    f"register_transformer must provide any of classes, metaclass, attr, detector"
                )

            for c in classes:
                assert inspect.isclass(
                    c
                ), f"register_transformer classes must be class, got {c}"

            if attr:
                assert isinstance(
                    attr, str
                ), f"register_transformer classes must be str, got {attr}"

            def detector(_cls):
                if classes:
                    if allow_subclasses:
                        if not issubclass(_cls, classes):
                            return False
                    else:
                        if _cls not in classes:
                            return False
                if metaclass:
                    if not isinstance(_cls, metaclass):
                        return False
                if attr and not hasattr(_cls, attr):
                    return False
                return True

        def decorator(f):
            if not self.validator(f):
                raise TypeError(f'Invalid register target: {f}, must pass <{self.validator}> validate')
            self._registry.insert(0, (detector, f, priority))
            if priority:
                self._registry.sort(key=lambda v: -v[2])
            return f

        # before runtime, type will be compiled and applied
        # if transformer is defined after the validator compiled
        # it will not take effect
        return decorator

    def resolve(self, t: type) -> Optional[Callable]:
        # resolve to it's subclass if both subclass and baseclass is provided
        # like Schema type will not resolve to dict
        if self.shortcut and hasattr(t, self.shortcut) and self.validator(getattr(t, self.shortcut)):
            # this type already got a callable transformer, do not resolve then
            return getattr(t, self.shortcut)
        if self.cache and t in self._cache:
            return self._cache[t]
        for detector, trans, priority in self._registry:
            try:
                if detector(t):
                    if self.cache:
                        self._cache[t] = trans
                    return trans
            except (TypeError, ValueError):
                continue
        if self.base:
            # default to base
            return self.base.resolve(t)
        return self.default



import inspect
import math
import warnings
import random
from decimal import Decimal
from enum import Enum, EnumMeta
from ..utils.datastructures import unprovided
from ..parser.field import ParserField
from ..parser.base import BaseParser
from ..parser.rule import LogicalType, Rule, SEQ_TYPES, MAP_TYPES
from uuid import UUID
from datetime import date, datetime, timedelta, time
from typing import Type

VALUE_TYPES = (str, int, float, Decimal, datetime, date, time, timedelta)


def get_example_from_json_schema(schema):
    pass


def get_example_from(t: type):
    if t == type(None):
        return None

    if t == bool:
        return random.choice([True, False])

    if t == UUID:
        import uuid
        return uuid.uuid4()

    if isinstance(t, EnumMeta):
        val = random.choice(t.__members__.values())     # noqa
        return t(val.value)

    parser = getattr(t, '__parser__', None)
    if isinstance(parser, BaseParser):
        return t(**get_example_from_parser(parser))

    if inspect.isclass(t) and issubclass(t, Rule):
        return get_example_from_rule(t)

    if isinstance(t, LogicalType):
        return t.get_example()

    return t()


def get_example_from_field(field: ParserField):
    if not unprovided(field.field.example):
        return field.field.example
    return get_example_from(field.type)


def get_example_from_parser(self):
    data = {}
    for name, field in self.fields.items():
        try:
            example = get_example_from_field(field)
        except Exception as e:
            warnings.warn(f'{self.obj}: generate example for field: [{repr(name)}] failed with error: {e}')
            continue
        data[name] = example
    return data


def get_example_from_rule(cls: Type[Rule]):
    """
    If example is forced and there is unsolvable rules (validator / converter) and no example provided
    will prompt error to ask provide example
    """
    if hasattr(cls, 'const'):
        return cls.const

    if hasattr(cls, 'enum'):
        return random.choice(cls.enum)

    if hasattr(cls, 'regex'):
        try:
            import exrex    # noqa
            return exrex.getone(cls.regex)
        except (ModuleNotFoundError, AttributeError):
            pass

    t = cls.__origin__

    if t in SEQ_TYPES:
        if cls.__args__:
            values = []
            if cls.__ellipsis_args__:
                # tuple
                for arg in cls.__args__:
                    values.append(
                        get_example_from(arg)
                    )
            else:
                values.append(
                    get_example_from(cls.__args__[0])
                )
            return t(values)

    if t in MAP_TYPES:
        if cls.__args__:
            values = {}
            key_type = cls.__args__[0]
            val_type = None
            if len(cls.__args__) > 1:
                val_type = cls.__args__[1]
            key = get_example_from(key_type)
            val = get_example_from(val_type) if val_type else None
            values[key] = val
            return t(values)

    if t not in VALUE_TYPES:
        return get_example_from(t)

    multi_of = getattr(cls, 'multiple_of', None)

    length = getattr(cls, 'length', None)
    min_length = getattr(cls, 'min_length', None)
    max_length = getattr(cls, 'max_length', None)

    round_v = getattr(cls, 'decimal_places', None)

    ge = getattr(cls, 'ge', None)
    gt = getattr(cls, 'gt', None)
    le = getattr(cls, 'le', None)
    lt = getattr(cls, 'lt', None)
    min_value = getattr(cls, 'ge', getattr(cls, 'gt', None))
    max_value = getattr(cls, 'le', getattr(cls, 'lt', None))

    if min_value is None:
        if max_value is None:
            if t == datetime:
                return datetime.now()
            elif t == date:
                return datetime.now().date()
            elif t == time:
                return datetime.now().time()
            elif t == timedelta:
                v = datetime.now().time()
                return timedelta(hours=v.hour, minutes=v.minute, seconds=v.second, microseconds=v.microsecond)
            else:
                if isinstance(multi_of, (int, float)):
                    return multi_of * random.randint(1, 10)

                val_length = length
                if val_length is None:
                    min_len = min_length or 0
                    max_len = max_length or min_len + 10
                    val_length = int(min_len + (max_len - min_len) * random.random())

                if not val_length:
                    return t()

                if t == str:
                    import string
                    return ''.join(random.sample(string.digits + string.ascii_letters, val_length))

                elif t in (float, Decimal):
                    if round_v and round_v > 0:
                        val_length = max(val_length - round_v - 1, 1)
                    val = random.randint(10 ** (val_length - 1), 10 ** val_length - 1) + random.random()
                    if round_v is not None:
                        val = round(val, round_v)
                    while len(str(val)) < val_length:
                        val = float(str(val) + '1')
                    return t(val)

                elif t == int:
                    return random.randint(10 ** (val_length - 1), 10 ** val_length - 1)
        else:
            if t in (int, float, Decimal):
                if isinstance(multi_of, (int, float)) and multi_of:
                    times = int(max_value / multi_of)
                    return multi_of * random.randint(max(1, times - 3), times)

                min_value = (max_value / 2) if max_value > 0 else max_value * 2
            elif t in (datetime, date):
                min_value = max_value - timedelta(days=1)   # noqa
            elif t == timedelta:
                min_value = min(timedelta(), max_value - timedelta(days=1))     # noqa
            elif t == time:
                min_value = time()

    elif max_value is None:
        if t in (int, float, Decimal):
            if isinstance(multi_of, (int, float)) and multi_of:
                times = math.ceil(min_value / multi_of)
                return multi_of * random.randint(times, times + 5)

            max_value = (min_value * 2) if min_value > 0 else min_value / 2
        elif t in (datetime, date, timedelta):
            max_value = min_value + timedelta(days=1)  # noqa
        elif t == time:
            max_value = time(23, 59, 59)

    if max_value is not None and min_value is not None:
        try:
            delta = max_value - min_value
            if isinstance(delta, int):
                if isinstance(multi_of, (int, float)) and multi_of:
                    max_times = int(max_value / multi_of)
                    min_times = math.ceil(min_value / multi_of)
                    return multi_of * random.randint(min_times, max_times)

                v = min_value + random.randint(0, delta)
                if v == gt:
                    v = v + 1
                elif v == lt:
                    v = v - 1
            else:
                v = min_value + delta * random.random()

            v = t(v)
            if round_v is not None:
                v = round(v, round_v)
            return v
        except TypeError:
            # like str
            seq = []
            if le:
                seq.append(le)
            if ge:
                seq.append(ge)
            if seq:
                return random.choice(seq)
            if isinstance(min_value, str):
                return min_value + '_' + str(random.randint(0, 9))
    return t()



import inspect
import warnings
from collections.abc import Mapping
from functools import partial
from typing import Callable, Dict, Type, TypeVar

from ..utils import exceptions as exc
from ..utils.compat import is_classvar, is_final
from ..utils.datastructures import unprovided
from ..utils.functional import pop
from ..utils.transform import TypeTransformer
from .base import BaseParser
from .field import ParserField
from .func import FunctionParser
from .options import Options, RuntimeContext

T = TypeVar("T")

__all__ = ["ClassParser", "init_dataclass"]


class ClassParser(BaseParser):
    # IGNORE_ATTR_TYPES = (staticmethod, classmethod, FunctionType, type)
    # if these type not having annotation, we will not recognize them as field
    function_parser_cls = FunctionParser
    fields: Dict[str, ParserField]

    def __init__(self, obj, *args, **kwargs):
        if not inspect.isclass(obj):
            raise TypeError(f"{self.__class__}: object need to be a class, got {obj}")
        super().__init__(obj, *args, **kwargs)
        self.init_parser = None

    @property
    def in_out_identical(self):
        for val in self.fields.values():
            if val.no_input is True and val.no_output is True:
                # ignored
                continue
            if val.no_input or val.no_output:
                return False
        return True

    def setup(self):
        self.generate_from_bases()
        super().setup()

    def validate_class_field_name(self, name: str):
        if not self.validate_field_name(name):
            return False
        for base in self.obj.__bases__:
            if base is object:
                continue
            annotations = getattr(base, "__annotations__", None)
            if annotations:
                # maybe object
                annotation = annotations.get(name)
                if annotation:
                    if is_final(annotation):
                        raise TypeError(
                            f"field: {repr(name)} was declared as Final in {base}, "
                            f"so {self.obj} cannot annotate it again"
                        )

            attr = getattr(base, name, None)
            if self.is_class_internals(
                attr, attname=name, class_qualname=base.__qualname__
            ):
                raise TypeError(
                    f"field: {repr(name)} was declared in {base}, "
                    f"so {self.obj} cannot annotate it as a field"
                )
        return True

    @classmethod
    def is_class_internals(cls, attr, attname: str, class_qualname: str = None):
        if isinstance(attr, (staticmethod, classmethod)):
            return True
        if inspect.ismethoddescriptor(attr):
            # like method_descriptor
            return True
        qualname: str = getattr(attr, "__qualname__", None)
        name: str = getattr(attr, "__name__", None)
        if name and qualname:
            if not class_qualname:
                # loosely check
                return attname == name and "." in qualname

            if attname == name and qualname.startswith(f"{class_qualname}."):
                return True
        return False

    @property
    def globals(self) -> dict:
        dic = dict(super().globals)
        # for the self-reference model in the function (not in global vars)
        # dic.setdefault(self.obj.__name__, self.obj)
        name = self.obj.__name__
        current_obj = dic.get(name)
        if current_obj:
            if getattr(current_obj, '__qualname__', None) != getattr(self.obj, '__qualname__', None):
                warnings.warn(f'Parser object: {self.obj} got conflict object: {current_obj} '
                              f'with same name: {repr(name)}, it may affect the ForwardRef resolve')
        dic[name] = self.obj
        # !IMPORTANT: we need to override __name__ for current obj
        # cause in the locals, same name may be the different object, we should be careful about that
        return dic

    def generate_fields(self):
        exclude_vars = self.exclude_vars
        fields = []

        annotations = self.obj.__dict__.get("__annotations__", {})
        # get annotations from __dict__
        # because if base has annotations and sub does not
        # it will directly use the annotations attr of base's
        global_vars = self.globals

        for key, annotation in annotations.items():
            if (
                not self.validate_class_field_name(key)
                or is_classvar(annotation)
                # or is_final(annotation)
            ):
                exclude_vars.add(key)
                continue
            default = self.obj.__dict__.get(key, unprovided)
            if annotation is None:
                # a: None
                # a: Optional[None]
                # a: Union[None]
                annotation = type(None)
                # to make a difference to annotation=None

            try:
                field = self.parser_field_cls.generate(
                    attname=key,
                    annotation=annotation,
                    default=default,
                    global_vars=global_vars,
                    forward_refs=self.forward_refs,
                    options=self.options,
                    force_clear_refs=self.is_local,
                    **self.kwargs
                )
            except Exception as e:
                raise exc.ConfigError(f'{self.name}: parse field [{repr(key)}] failed with error: {e}')

            fields.append(field)

        for key, attr in self.obj.__dict__.items():
            if key in annotations:
                continue
            if (
                # if this attr is a field in bases, this means to exclude this field in current class
                # otherwise this attr declared that this field is never take from input
                # or isinstance(attr, property)
                self.is_class_internals(attr, attname=key, class_qualname=self.obj_name)
                # or isinstance(attr, self.IGNORE_ATTR_TYPES)
                # check class field name at last
                # because this will check bases internals trying to find illegal override
                or not self.validate_class_field_name(key)
            ):
                # key is not consider a valid field
                # and not
                exclude_vars.add(key)
                continue
            if key in exclude_vars:
                continue
            if attr is Ellipsis and key in self.fields:
                # class base(Schema):
                #     f: int
                # class sub(base):
                #     f = ...
                # means that sub schema has dropped field [f] (not declaring the annotation)
                self.fields.pop(key)
                continue

            try:
                field = self.parser_field_cls.generate(
                    attname=key,
                    annotation=None,
                    default=attr,
                    global_vars=global_vars,
                    forward_refs=self.forward_refs,
                    options=self.options,
                    force_clear_refs=self.is_local,
                    **self.kwargs
                )
            except Exception as e:
                raise exc.ConfigError(f'{self.name}: parse field [{repr(key)}] failed with error: {e}')

            fields.append(field)

        field_map = {}
        for field in fields:
            name = field.name

            if field.is_case_insensitive(self.options):
                name = name.lower()

            if name in field_map:
                raise exc.ConfigError(
                    f"{self.obj}: field name: {repr(name)} conflicted at "
                    f"{field}, {field_map[name]}",
                    obj=self.obj,
                    field=name,
                )
            field_map[name] = field

        self.fields.update(field_map)

    def generate_from_bases(self):
        fields = {}
        alias_map = {}
        attr_alias_map = {}
        case_insensitive_names = set()
        exclude_vars = set()
        # option_list = []

        for base in reversed(self.obj.__bases__):  # according to MRO
            if not isinstance(base, type(self.obj)) or base is object:
                continue
            parser = self.apply_for(base)  # should use cache
            # if not parser.options.vacuum:
            #     option_list.append(parser.options)

            fields.update(parser.fields)

            exclude_vars.update(parser.exclude_vars)
            alias_map.update(parser.field_alias_map)
            attr_alias_map.update(parser.attr_alias_map)
            case_insensitive_names.update(parser.case_insensitive_names)

        # cls_options = self.options  # add current cls options
        # if cls_options:
        #     option_list.append(cls_options)
        #
        # self.options = self.options_cls.generate_from(*option_list)

        self.fields = fields
        self.exclude_vars = exclude_vars
        self.field_alias_map = alias_map
        self.attr_alias_map = attr_alias_map
        self.case_insensitive_names = case_insensitive_names

    def make_setter(self, field: ParserField, post_setattr=None):
        def setter(_obj_self: object, value):
            if self.options.immutable or field.immutable:
                raise exc.UpdateError(
                    f"{self.name}: "
                    f"Attempt to set immutable attribute: [{repr(field.attname)}]"
                )

            context = self.options.make_context(_obj_self.__class__, force_error=True)
            value = field.parse_value(value, context=context)
            _obj_self.__dict__[field.attname] = value
            if callable(post_setattr):
                post_setattr(_obj_self, field, value, context)

        return setter

    def make_deleter(self, field: ParserField, post_delattr=None):
        def deleter(_obj_self: object):
            if self.options.immutable or field.immutable:
                raise exc.DeleteError(
                    f"{self.name}: "
                    f"Attempt to set immutable attribute: [{repr(field.attname)}]"
                )

            context = self.options.make_context(_obj_self.__class__, force_error=True)
            if field.is_required(context.options):
                raise exc.DeleteError(
                    f"{self.name}: Attempt to delete required schema key: {repr(field.attname)}"
                )

            if field.attname not in _obj_self.__dict__:
                raise exc.DeleteError(
                    f"{self.name}: Attempt to delete nonexistent key: {repr(field.attname)}"
                )

            _obj_self.__dict__.pop(field.attname)

            if callable(post_delattr):
                post_delattr(_obj_self, field, context)

        return deleter

    def make_getter(self, field: ParserField):
        def getter(_obj_self: object):
            if field.attname not in _obj_self.__dict__:
                raise AttributeError(
                    f"{self.name}: {repr(field.attname)} not provided in schema"
                )
            return _obj_self.__dict__[field.attname]

        return getter

    def assign_properties(
        self,
        getter: Callable = None,
        setter: Callable = None,
        deleter: Callable = None,
        post_setattr: Callable = None,
        post_delattr: Callable = None,
    ):

        for key, field in self.fields.items():
            if field.property:
                continue

            if getter:
                field_getter = partial(getter, field=field)
            else:
                field_getter = self.make_getter(field)
            if setter:
                field_setter = partial(setter, field=field)
            else:
                field_setter = self.make_setter(field, post_setattr=post_setattr)
            if deleter:
                field_deleter = partial(deleter, field=field)
            else:
                field_deleter = self.make_deleter(field, post_delattr=post_delattr)

            for f in (field_getter, field_setter, field_deleter):
                f.__name__ = field.attname

            prop = property(fget=field_getter, fset=field_setter, fdel=field_deleter)
            # prop.__field__ = field        # cannot set attribute to @property
            setattr(self.obj, field.attname, prop)

    def get_parser(self, obj_self: object):
        if self.obj == obj_self.__class__:
            return self
        return self.resolve_parser(obj_self.__class__)

    def _make_method(self, func: Callable, name: str = None):
        if name:
            func.__name__ = name
        else:
            name = func.__name__

        if name in self.obj.__dict__:
            # already declared
            return False
        attr_func = getattr(self.obj, name, None)
        if hasattr(attr_func, "__parser__"):
            # already inherited
            return False
        func.__parser__ = self
        setattr(self.obj, name, func)
        return True

    def make_contains(self, output_only: bool = False):
        def __contains__(_obj_self, item: str):
            parser = self.get_parser(_obj_self)
            field = parser.get_field(item)
            if not field:
                return False
            if field.attname not in _obj_self.__dict__:
                return False
            if not output_only:
                return True
            if field.is_no_output(
                _obj_self.__dict__[field.attname], options=parser.options
            ):
                return False
            return True

        self._make_method(__contains__)

    def make_eq(self):
        def __eq__(_obj_self, other):
            if not isinstance(other, _obj_self.__class__):
                return False
            if _obj_self.__dict__ == other.__dict__:
                return True
            self_dict = dict(_obj_self.__dict__)
            pop(self_dict, "__context__")
            other_dict = dict(other.__dict__)
            pop(other_dict, "__context__")
            return self_dict == other_dict

        self._make_method(__eq__)

    def make_repr(self, ignore_str: bool = False):
        def __repr__(_obj_self):
            parser = self.get_parser(_obj_self)
            items = []
            for key, val in _obj_self.__dict__.items():
                field = parser.get_field(key)
                if not field:
                    continue
                items.append(f"{field.attname}={field.repr_value(val)}")
            values = ", ".join(items)
            return f"{parser.name}({values})"

        self._make_method(__repr__)

        if not ignore_str:

            def __str__(_obj_self):
                return _obj_self.__repr__()

            self._make_method(__str__)

    def set_attributes(
        self,
        values: dict,
        instance: object,
        options: Options,
    ):

        for key, value in list(values.items()):
            field = self.get_field(key)
            attname = key
            if field:
                if field.is_no_output(values[key], options=options):
                    values.pop(key)
                if field.property:
                    try:
                        field.property.fset(
                            instance, values[key]
                        )  # call the original setter
                        # setattr(instance, field.attname, values[key])
                    except Exception as e:
                        error_option = field.get_on_error(options)
                        msg = f"@property: {repr(field.attname)} assign failed with error: {e}"
                        if error_option == options.THROW:
                            raise e.__class__(msg) from e
                        else:
                            warnings.warn(msg)
                    continue
                attname = field.attname

            # TODO: it seems redundant for Schema, so we just use it as a fallback for now
            # and work on it later if something went wrong
            instance.__dict__[attname] = value
            # set to __dict__ no matter field (maybe addition=True)

    def make_context(self, context=None, force_error: bool = False):
        return self.options.make_context(self.obj, context=context, force_error=force_error)

    def make_init(
        self,
        # init_super: bool = False,
        # allow_runtime: bool = False,
        # set_attributes: bool = True,
        # coerce_property: bool = False,
        no_parse: bool = False,
        post_init: Callable = None,
    ):

        init_func = getattr(self.obj, "__init__", None)

        init_parser = self.resolve_parser(init_func)
        if init_parser:
            # if init_func is already decorated like a Wrapper
            # we do not touch it either
            # case1: user use @utype.parse over the __init__ function
            # case2: base ClassParser has assigned the wrapped init with __parser__ attribute
            self.init_parser = init_parser
            return

        if not inspect.isfunction(init_func) or self.function_parser_cls.function_pass(
            init_func
        ):
            # if __init__ is declared but passed, we still make a new one

            def __init__(_obj_self, _d: dict = None, **kwargs):
                parser = self.get_parser(_obj_self)

                context = getattr(_obj_self, "__context__", None)
                if not isinstance(context, RuntimeContext):
                    context: RuntimeContext = parser.make_context()

                if isinstance(_d, dict):
                    kwargs.update(_d)

                if no_parse:
                    values = kwargs
                else:
                    values = parser(kwargs, context=context)

                parser.set_attributes(values, _obj_self, options=context.options)

                if post_init:
                    post_init(_obj_self, values, context)

            __init__.__parser__ = self
        else:
            if not no_parse:
                self.init_parser = self.function_parser_cls.apply_for(init_func)
                if self.init_parser.pos_only_keys:
                    raise exc.ConfigError(
                        f"{self.obj}: positional only keys: {self.init_parser.pos_only_keys} "
                        f"is not allowed for dataclasses __init__",
                        obj=self.obj,
                    )
                if self.init_parser.pos_var:
                    raise exc.ConfigError(
                        f"{self.obj}: positional var: {self.init_parser.pos_var} "
                        f"is not allowed for dataclasses __init__",
                        obj=self.obj,
                    )

                __init__ = self.init_parser.wrap(parse_params=True, parse_result=False)
                __init__.__parser__ = self
                # wrapped function is not as same as parse.obj
            else:
                __init__ = init_func

        setattr(self.obj, "__init__", __init__)
        # self.obj.__dict__['__init__'] = __init__
        # the INPUT parser
        # we do not merge fields or options here
        # each part does there job
        # init just parse data as it declared and take it to initialize the class
        # class T(Schema):
        #     mul: int
        #     def __init__(self, a: float, b: int):
        #         super().__init__(mul=a * b)
        # we will make init_parser the "INPUT" parser

        return __init__

    @property
    def schema_annotations(self):
        # this is meant to be extended and override
        # if the result is not None, it will become the x-annotation of the JSON schema output
        data = dict()
        if self.options.mode:
            data.update(mode=self.options.mode)
        if self.options.case_insensitive:
            data.update(case_insensitive=self.options.case_insensitive)
        return data


def init_dataclass(
    cls: Type[T], data, options: Options = None, context: RuntimeContext = None
) -> T:
    parser: ClassParser = getattr(cls, "__parser__", None)
    if not isinstance(parser, ClassParser):
        raise exc.TypeMismatchError(f"Invalid dataclass: {cls}")

    if options:
        new_context: RuntimeContext = options.make_context(cls, context=context)
    else:
        new_context: RuntimeContext = parser.make_context(context=context)

    transformer = new_context.transformer

    try:
        if not isinstance(data, Mapping):
            # {} dict instance is an instance of Mapping too
            if transformer.no_explicit_cast:
                raise TypeError(
                    f"invalid input type for {cls}, should be dict or Mapping"
                )
            else:
                data = transformer.to_dict(data)

        if new_context.options.cast_keyword_str:
            _data = {}
            for key, val in data.items():
                if not isinstance(key, str):
                    key = transformer.to_str(key)
                _data[key] = val
            data = _data
    except Exception as e:
        raise exc.ParseError(type=cls, value=data, origin_exc=e) from e

    inst = cls.__new__(cls)
    inst.__context__ = new_context

    # if parser.init_parser:
    cls.__init__(inst, **data)

    return inst


@TypeTransformer.registry.register(
    attr="__parser__",
    detector=lambda cls: isinstance(getattr(cls, "__parser__", None), ClassParser),
)
def transform_dataclass(transformer: TypeTransformer, data, cls):
    if isinstance(data, (list, tuple)) and not transformer.options.no_explicit_cast:
        if data:
            if transformer.options.no_data_loss and len(data) > 1:
                raise TypeError
            data = data[0]
            # otherwise the data will become dict then fill the dataclass
            if type(data) == cls:
                return data

    if transformer.options.allow_subclasses:
        if isinstance(data, cls):
            # subclass
            return data

    return init_dataclass(cls, data, context=transformer.context)



from .decorator import apply, dataclass, handle, parse, raw
from .parser.field import Field, Param
from .parser.options import Options
from .parser.rule import Lax, Rule
from .schema import DataClass, LogicalMeta, Schema
from .utils import exceptions as exc
from .utils.encode import register_encoder, JSONEncoder
from .utils.transform import (TypeTransformer, type_transform)
from .utils.datastructures import unprovided
from .specs.json_schema import JsonSchemaGenerator

register_transformer = TypeTransformer.registry.register


VERSION = (0, 5, 0, None)


def _get_version():
    pre_release = VERSION[3] if len(VERSION) > 3 else ""
    version = ".".join([str(v) for v in VERSION[:3]])
    if pre_release:
        version += f"-{pre_release}"
    return version


__version__ = _get_version()


def version_info() -> str:
    import platform
    import sys
    from pathlib import Path

    info = {
        'utype version': __version__,
        'installed path': Path(__file__).resolve().parent,
        'python version': sys.version,
        'platform': platform.platform(),
    }
    return '\n'.join('{:>30} {}'.format(k + ':', str(v).replace('\n', ' ')) for k, v in info.items())



"""
Python type hint standard may vary from versions
"""
import sys
import typing
from typing import (Any, ClassVar, ForwardRef, Optional, Tuple,  # type: ignore
                    Type, Union)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

try:
    from typing import Final
except ImportError:
    from typing_extensions import Final

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated


__all__ = [
    "get_origin",
    "get_args",
    'Literal',
    'Final',
    "ForwardRef",
    "Annotated",
    "is_final",
    "is_union",
    "is_classvar",
    "is_annotated",
    "evaluate_forward_ref",
    'JSON_TYPES'
]

if sys.version_info < (3, 8):

    def get_origin(t) -> Optional[Type[Any]]:
        return getattr(t, "__origin__", None)

    def get_args(t) -> Tuple[Type[Any], ...]:
        return getattr(t, "__args__", ())

else:
    from typing import get_args as _typing_get_args
    from typing import get_origin as _typing_get_origin

    def get_origin(t) -> Optional[Type[Any]]:
        return _typing_get_origin(t) or getattr(t, "__origin__", None)

    def get_args(t) -> Tuple[Type[Any], ...]:
        args = getattr(t, "__args__", ())
        if args == ():
            # typing.get_args(typing.Callable) will throw an error
            return args
        return _typing_get_args(t) or args


try:
    from typing import ForwardRef  # type: ignore

    def evaluate_forward_ref(ref: ForwardRef, globalns: Any, localns: Any):
        return typing._eval_type(ref, globalns, localns)  # noqa

except ImportError:
    # python 3.6
    from typing import _ForwardRef as ForwardRef  # noqa

    def evaluate_forward_ref(ref: ForwardRef, globalns: Any, localns: Any):
        return ref._eval_type(globalns, localns)  # noqa


if sys.version_info < (3, 10):

    def is_union(tp: Optional[Type[Any]]) -> bool:
        return tp is Union

else:
    import types

    def is_union(tp: Optional[Type[Any]]) -> bool:
        return tp is Union or tp is types.UnionType  # noqa


def _check_classvar(v) -> bool:
    return type(v) == type(ClassVar) and (
        sys.version_info < (3, 7) or getattr(v, "_name", None) == "ClassVar"
    )


def _check_final(v) -> bool:
    return type(v) == type(Final) and (
        sys.version_info < (3, 7) or getattr(v, "_name", None) == "Final"
    )


def is_classvar(ann_type) -> bool:
    return _check_classvar(ann_type) or _check_classvar(
        getattr(ann_type, "__origin__", None)
    )


_AnnotatedTypeNames = {'AnnotatedMeta', '_AnnotatedAlias'}


def is_annotated(ann_type) -> bool:
    # duck type check for Annotated, name in AnnotatedTypeNames and have __origin__ attribute
    return type(ann_type).__name__ in _AnnotatedTypeNames and getattr(ann_type, '__origin__', None)


def is_final(ann_type) -> bool:
    return _check_final(ann_type) or _check_final(getattr(ann_type, "__origin__", None))


ATOM_TYPES = (str, int, bool, float, type(None))
JSON_TYPES = (*ATOM_TYPES, list, dict)
# types thar can directly dump to json
# COMMON_TYPES = (*JSON_TYPES, set, tuple, bytes, *VENDOR_TYPES)



# ERROR for handling parse
from typing import Any, List, Set, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from ..parser.field import ParserField


class ConfigError(SyntaxError):
    def __init__(self, msg="", obj=None, params: dict = None, field: str = None):
        super().__init__(msg)
        self.obj = obj
        self.params = params
        self.field = field


class FieldError(AttributeError, KeyError):
    def __init__(
        self,
        msg: str = None,
        *,
        field=None,
        origin_exc: Exception = None,
    ):
        self.msg = msg
        self.field = field
        self.origin = origin_exc

        super().__init__(msg)


class UpdateError(FieldError):
    pass


class DeleteError(FieldError):
    pass


class ParseError(TypeError, ValueError):
    def __init__(
        self,
        msg: str = None,
        *,
        value: Any = None,
        type: type = None,
        item: Union[
            int, str
        ] = None,  # like key in the object or index in seq to indentify value
        field: 'ParserField' = None,  # no field can means it's additional field
        routes: list = None,
        origin_exc: Exception = None,
    ):
        if not msg and origin_exc:
            msg = str(origin_exc)
        self.msg = msg
        self.origin_exc = origin_exc
        self.value = value
        self.type = type
        self.item = item
        self.field = field
        self.routes = routes
        super().__init__(self.formatted_message)

    @property
    def formatted_message(self):
        msg = self.msg
        if self.item:
            msg = f"parse item: [{repr(self.item)}] failed: {msg}"
        return msg

    def get_detail(self) -> dict:
        from ..specs.json_schema import JsonSchemaGenerator
        origin = None
        if self.origin_exc:
            if isinstance(self.origin_exc, ParseError) and self.origin_exc.field:
                origin = self.origin_exc.get_detail()

        if origin:
            return {
                'name': self.field.name if self.field else None,
                'field': self.field.field.__class__.__name__ if self.field else None,
                'origin': origin
            }
        return {
            'name': self.field.name if self.field else None,
            'value': self.value,
            'field': self.field.field.__class__.__name__ if self.field else None,
            'schema': JsonSchemaGenerator(self.type)() if self.type else None,
            'msg': self.msg,
        }


class TypeMismatchError(ParseError):
    @property
    def formatted_message(self):
        msg = f"type: {self.type} is unrecognized and forbid to auto-init"
        if self.msg:
            msg += f": {self.msg}"
        return msg


class InvalidInstance(TypeMismatchError):
    @property
    def formatted_message(self):
        msg = f"invalid class instance: {self.value} for {self.type}"
        if self.msg:
            msg += f": {self.msg}"
        return msg


class InvalidSubclass(TypeMismatchError):
    @property
    def formatted_message(self):
        msg = f"invalid subclass: {self.value} for {self.type}"
        if self.msg:
            msg += f": {self.msg}"
        return msg


class DiscriminatorMismatchError(ParseError):
    def __init__(self, discriminator: str, discriminator_value=None, **kwargs):
        self.discriminator = discriminator
        self.discriminator_value = discriminator_value
        super().__init__(**kwargs)


class ConstraintError(ParseError):
    def __init__(
        self,
        msg: str = None,
        *,
        value: Any = None,
        type: type = None,
        item: Union[
            int, str
        ] = None,  # like key in the object or index in seq to indentify value
        constraint: str = None,  # failed constraint
        constraint_value: Any = None,  # failed constraint value
        origin_exc: Exception = None,
    ):
        self.constraint = constraint
        self.constraint_value = constraint_value
        super().__init__(msg, value=value, type=type, item=item, origin_exc=origin_exc)

    @property
    def formatted_message(self):
        if self.constraint:
            msg = f"Constraint: <{self.constraint}>: {repr(self.constraint_value)} violated"
            if self.msg:
                msg += f": {self.msg}"
            return msg
        return self.msg


class ExceedError(ParseError):
    # a key has excess the dict template and allow_excess=False in options
    # def __init__(
    #     self, msg: str = None, excess_items: Union[list, set] = None, **kwargs
    # ):
    #     self.excess_items = excess_items
    #     super().__init__(msg, **kwargs)

    @property
    def formatted_message(self):
        msg = f"parse item: [{repr(self.item)}] exceeded"
        if self.msg:
            msg += f": {self.msg}"
        return msg


class TupleExceedError(ExceedError):
    pass


class AliasConflictError(ParseError):
    pass


# class ItemsExceedError(ExceedError):
#     pass


class DepthExceedError(ParseError):
    def __init__(
        self,
        msg: str = None,
        max_depth: int = None,
        depth: int = None,
        **kwargs,
    ):
        self.depth = depth
        self.max_depth = max_depth
        super().__init__(msg, **kwargs)

    @property
    def formatted_message(self):
        msg = f"max_depth: {self.max_depth} exceed: {self.depth}"
        if self.msg:
            msg += f": {self.msg}"
        return msg


class ParamsExceedError(ParseError):
    def __init__(
        self,
        msg: str = None,
        max_params: int = None,
        params_num: int = None,
        **kwargs,
    ):
        self.params_num = params_num
        self.max_params = max_params
        super().__init__(msg, **kwargs)

    @property
    def formatted_message(self):
        msg = f"max params num: {self.max_params} exceed: {self.params_num}"
        if self.msg:
            msg += f": {self.msg}"
        return msg


class ParamsLackError(ParseError):
    def __init__(
        self,
        msg: str = None,
        min_params: int = None,
        params_num: int = None,
        **kwargs,
    ):
        self.params_num = params_num
        self.min_params = min_params
        super().__init__(msg, **kwargs)

    @property
    def formatted_message(self):
        msg = f"min params num: {self.min_params} lacked: {self.params_num}"
        if self.msg:
            msg += f": {self.msg}"
        return msg


class AbsenceError(ParseError):
    @property
    def formatted_message(self):
        msg = f"required item: {repr(self.item)} is absence"
        if self.msg:
            msg += f": {self.msg}"
        return msg


class DependenciesAbsenceError(AbsenceError):
    def __init__(
        self, msg: str = None, absence_dependencies: Set[str] = None, **kwargs
    ):
        self.absence_dependencies = absence_dependencies
        super().__init__(msg, **kwargs)

    @property
    def formatted_message(self):
        msg = f"required dependencies: {self.absence_dependencies} is absence"
        if self.msg:
            msg += f": {self.msg}"
        return msg


class RecursionExceeded(ParseError, RecursionError):
    def __init__(self, msg: str = None, depth: int = None, **kwargs):
        super().__init__(msg, **kwargs)
        self.depth = depth


class TransformError(ParseError, TypeError):
    def __init__(self, msg: str = None, data_type: type = None, **kwargs):
        super().__init__(msg, **kwargs)
        self.data_type = data_type


class CollectedParseError(ParseError):
    def __init__(self, errors: List[ParseError]):
        self.errors = errors
        super().__init__(";\n".join([str(error) for error in errors]))

    def get_detail(self) -> list:
        errors = []
        for error in self.errors:
            if isinstance(error, ParseError):
                errors.append(error.get_detail())
            else:
                errors.append({
                    'msg': str(error),
                    'error': error.__class__.__name__
                })
        return errors


class NegateViolatedError(ParseError):
    pass


class OneOfViolatedError(ParseError):
    pass



import inspect
import warnings
from typing import Any, Callable, Iterable, Type, TypeVar, Union

from .parser.cls import ClassParser
from .parser.func import FunctionParser
from .parser.options import Options
from .parser.rule import Lax, Rule
from .utils import exceptions as exc
from .utils.datastructures import Unprovided, unprovided

T = TypeVar("T")
FUNC = TypeVar("FUNC")
CLS = TypeVar("CLS")


def raw(f: FUNC):
    parser = getattr(f, "__parser__", None)
    if isinstance(parser, FunctionParser):
        return parser.obj
    return f


def parse(
    f: T = None,
    # /,  compat 3.7
    *,
    parser_cls: Type[FunctionParser] = FunctionParser,
    options: Union[Options, Type[Options]] = None,
    no_cache: bool = False,
    # static: bool = False,
    ignore_params: bool = False,
    ignore_result: bool = False,
    eager: bool = False,
) -> Union[T, Callable[[T], T]]:
    if ignore_params and ignore_result:
        warnings.warn(
            f"@utype.parse: you turn off both params and result parse in @parse decorator,"
            f" which is basically meaningless..."
        )

    def decorator(func: T) -> T:
        if inspect.isclass(func):
            # class
            parser_cls.apply_class(
                func,
                options=options,
                no_cache=no_cache,
                ignore_result=ignore_result,
                ignore_params=ignore_params,
                eager=eager,
            )
            return func
        else:
            parser = parser_cls.apply_for(func, options=options, no_cache=no_cache)
            return parser.wrap(
                parse_params=not ignore_params,
                parse_result=not ignore_result,
                eager_parse=eager
                # first_reserve=not static
            )

    if f:
        return decorator(f)
    return decorator


def dataclass(
    obj: CLS = None,
    # /,  compat 3.7
    *,
    parser_cls: Type[ClassParser] = ClassParser,
    options: Union[Options, Type[Options]] = None,
    no_cache: bool = False,
    no_parse: bool = False,
    set_class_properties: bool = False,
    post_init: Callable = None,
    post_setattr: Callable = None,
    post_delattr: Callable = None,
    contains: bool = False,
    repr: bool = True,  # noqa
    eq: bool = False,
) -> Union[CLS, Callable[[CLS], CLS]]:
    def decorator(cls: CLS) -> CLS:
        parser = parser_cls.apply_for(cls, options=options, no_cache=no_cache)

        parser.make_init(
            # init_super=init_super,
            # allow_runtime=allow_runtime,
            # set_attributes=init_attributes,
            # coerce_property=init_properties,
            no_parse=no_parse,
            post_init=post_init,
        )
        if repr:
            parser.make_repr()
        if eq:
            parser.make_eq()
        if contains:
            parser.make_contains()
        if set_class_properties:
            parser.assign_properties(
                post_setattr=post_setattr, post_delattr=post_delattr
            )
        elif post_setattr or post_delattr:
            warnings.warn(
                f"@utype.dataclass received post_delattr / post_setattr "
                f'without "set_properties=True", these params won\'t take effect'
            )

        cls.__parser__ = parser
        # set the flag so that class parser can be quickly resolve
        # and transformer can be register

        return parser.obj

    if obj:
        return decorator(obj)
    return decorator


def apply(
    rule_cls: Type[Rule] = Rule,
    *,
    # init: bool = False,     # whether to override init
    # -- constraints:
    const: Any = unprovided,
    enum: Iterable = None,
    gt=None,
    ge=None,
    lt=None,
    le=None,
    regex: str = None,
    length: int = None,
    max_length: int = None,
    min_length: int = None,
    # number
    max_digits: int = None,
    decimal_places: int = None,
    round: int = None,
    multiple_of: int = None,
    # array
    contains: type = None,
    max_contains: int = None,
    min_contains: int = None,
    unique_items: bool = None,
):
    # if rule_cls:
    #     pass

    if round:
        if decimal_places and decimal_places not in (round, Lax(round)):
            raise exc.ConfigError(
                f"@apply round: {round} is a shortcut for decimal_places=Lax({round}), "
                f"but you specified a different decimal_places: {repr(decimal_places)}"
            )

        decimal_places = Lax(round)

    constraints = {
        k: v
        for k, v in dict(
            enum=enum,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            length=length,
            regex=regex,
            max_digits=max_digits,
            decimal_places=decimal_places,
            multiple_of=multiple_of,
            contains=contains,
            max_contains=max_contains,
            min_contains=min_contains,
            unique_items=unique_items,
        ).items()
        if v is not None
    }

    if not isinstance(const, Unprovided):
        constraints.update(const=const)

    def decorator(_type):
        cls = rule_cls.annotate(_type, constraints=constraints)
        cls.__name__ = getattr(_type, "__name__", cls.__name__)
        cls.__repr__ = getattr(_type, "__repr__", cls.__repr__)
        cls.__str__ = getattr(_type, "__str__", cls.__str__)
        cls.__applied__ = True
        # applied Rule only checks constraints if value is not the instance of the __origin__ type
        return cls
        #
        # if init:
        #     pass
        # else:
        #     @register_transformer(_type)
        #     def _transform_type(trans, value, t):
        #         return type_cls(value)
        #
        #     _transform_type.__name__ = f'_to_{getattr(_type, "__name__", str(_type))}'
        # return _type

    return decorator


def handle(*func_and_errors):
    # implement in next version
    pass



import warnings
from typing import Callable, List, Union

from .functional import multi


class CaseStyle:
    camelCase = "camel"
    PascalCase = "pascal"
    snake_case = "snake"
    kebab_case = "kebab"
    CAP_KEBAB_CASE = "cap_kebab"
    CAP_SNAKE_CASE = "cap_snake"


CASE_STYLES = [
    CaseStyle.camelCase,
    CaseStyle.PascalCase,
    CaseStyle.snake_case,
    CaseStyle.kebab_case,
    CaseStyle.CAP_KEBAB_CASE,
    CaseStyle.CAP_SNAKE_CASE,
]


class AliasGenerator:
    @classmethod
    def guess_style(cls, style: Union[str, Callable]):
        if callable(style):
            return style
        if not isinstance(style, str) or not style:
            return None
        style = style.lower()
        if CaseStyle.camelCase in style:
            return CaseStyle.camelCase
        if CaseStyle.snake_case in style:
            if "cap" in style:
                return CaseStyle.CAP_SNAKE_CASE
            return CaseStyle.snake_case
        if CaseStyle.kebab_case in style:
            if "cap" in style:
                return CaseStyle.CAP_KEBAB_CASE
            return CaseStyle.kebab_case
        if CaseStyle.PascalCase in style:
            return CaseStyle.PascalCase
        # guess val
        ans = "".join(filter(str.isalnum, style))
        if "_" in style:
            if ans.isupper():
                return CaseStyle.CAP_SNAKE_CASE
            return CaseStyle.snake_case
        if "-" in style:
            if ans.isupper():
                return CaseStyle.CAP_KEBAB_CASE
            return CaseStyle.kebab_case
        if ans.isupper():
            return CaseStyle.CAP_SNAKE_CASE
        if ans.islower():
            return CaseStyle.snake_case
        if ans[0].islower():
            return CaseStyle.camelCase
        return CaseStyle.PascalCase

    def __init__(self, generator: Union[str, Callable], allow_conflict: bool = False):
        self.generator = self.guess_style(generator)
        if not self.generator:
            raise ValueError(f"Invalid case style: {generator}")
        self.allow_conflict = allow_conflict
        self.func = None
        if callable(self.generator):
            self.func = self.generator
        elif isinstance(self.generator, str):
            self.func = getattr(self.__class__, self.generator, None)
        if not self.func:
            raise ValueError(f"Invalid case style: {generator}")

    def __call__(self, data):
        if not self.generator or not data:
            return data
        if multi(data):
            return [self(d) for d in data]
        elif isinstance(data, dict):
            result = {}
            if data.get("@"):
                return data
            for key, val in data.items():
                k = self(key)
                if not self.allow_conflict and k in result:
                    raise ValueError(f"Duplicate data key: {k}")
                if multi(k) and k:
                    k = list(k)[0]
                if not isinstance(k, str):
                    # invalid key, ignore
                    continue
                result[k] = val
            return result
        elif isinstance(data, str):
            try:
                return self.func(data)
            except Exception as e:
                warnings.warn(
                    f"apply field transformer failed with error: {e}, ignoring..."
                )
                return data
        return data

    @classmethod
    def pascal(cls, val: str):
        if not val:
            return ""
        split = None
        if "_" in val:
            # guess type: snake / cap_snake
            split = "_"
        elif "-" in val:
            split = "-"
        if split:
            val = "".join(word.capitalize() for word in val.split(split))
        if not val.isalnum():
            return val
        # val = ''.join(filter(str.isalnum, val))
        if split:
            return val
        if val.islower():
            return val.capitalize()
        if val.isupper():
            return val.capitalize()
        if val[0].islower():
            # guess is PascalCase
            return val[0].upper() + val[1:]
        return val

    @classmethod
    def snake(cls, val: str):
        if not val:
            return ""
        if "_" in val:
            # guess type: snake / cap_snake
            return val.lower()
        elif "-" in val:
            return val.replace("-", "_").lower()

        if not val.isalnum():
            return val

        # val = ''.join(filter(str.isalnum, val))

        if val.islower():
            return val
        if val.isupper():
            return val.lower()

        s = ""
        for i, c in enumerate(val):
            if c.isupper():
                if i:
                    # not for first upper case
                    s += "_"
                c = c.lower()
            s += c

        return s

    @classmethod
    def camel(cls, val: str):
        val = cls.pascal(val)
        return val[0].lower() + val[1:]

    @classmethod
    def cap_snake(cls, val: str):
        return cls.snake(val).upper()

    @classmethod
    def kebab(cls, val: str):
        return cls.snake(val).replace("_", "-")

    @classmethod
    def cap_kebab(cls, val: str):
        return cls.cap_snake(val).replace("_", "-")

    @classmethod
    def generate_aliases(
        cls, val: str, generator: Union[str, Callable, List[str], bool] = "*"
    ):
        if not generator:
            return []
        if generator == "*" or generator is True:
            generator = CASE_STYLES
        elif not multi(generator):
            generator = [generator]
        aliases = []

        def _validator(v):
            return isinstance(v, str) and v and v != val

        for g in generator:
            res = cls(generator=g)(val)
            if multi(res):
                for r in res:
                    if _validator(r):
                        aliases.append(r)
            elif _validator(res):
                aliases.append(res)
        return aliases



import inspect
import sys
import warnings
from typing import Callable, Dict, List, Optional, Set, Tuple, Type, Union

from ..utils import exceptions as exc
from ..utils.compat import *
from ..utils.functional import is_local_var, get_obj_name
from ..utils.datastructures import cached_property, unprovided
from .field import ParserField
from .options import Options, RuntimeContext
from .rule import resolve_forward_type

__parsers__ = {}


class BaseParser:
    options_cls = Options
    parser_field_cls = ParserField

    # DEFAULT_EXCLUDE_VARS = {"__options__", "__class__"}

    @classmethod
    def resolve_parser(cls, obj):
        parser = getattr(obj, "__dict__", {}).get("__parser__")
        # get to the __dict__ first
        if isinstance(parser, cls):
            return parser
        if not is_local_var(obj):
            # return None
            global __parsers__
            if obj in __parsers__:
                return __parsers__[obj]
        parser = getattr(obj, "__parser__", None)
        if isinstance(parser, cls):
            return parser
        return None

    @classmethod
    def apply_for(cls, obj, no_cache: bool = False, options=None, **kwargs) -> "BaseParser":
        if isinstance(obj, cls):
            return obj
        parser = getattr(obj, "__dict__", {}).get("__parser__")
        # do not use getattr(obj, '__parser__')
        # since it may get to the base class
        if isinstance(parser, cls):
            return parser
        if is_local_var(obj):
            no_cache = True
        global __parsers__
        # key = (cls, obj)
        key = obj
        if not options:
            options = getattr(obj, '__options__', None)
        if not no_cache and key in __parsers__:
            cached: "BaseParser" = __parsers__[key]
            # if options is not identical, make a new one
            if not options or options == cached.options:
                return cached
        inst = cls(obj, options=options, **kwargs)      # noqa
        if not no_cache:
            __parsers__[key] = inst
        return inst

    def __init__(self, obj, options: Options = None):
        self.obj = obj
        self.init_kwargs = {"options": options}
        self.options: Options = self.options_cls.generate_from(options)

        self.forward_refs: Dict[
            str, Tuple[ForwardRef, dict]
        ] = {}  # store unresolved ref
        self.fields: Dict[str, ParserField] = {}
        self.exclude_vars: Set[str] = set()
        # these data structures are designed to speed up the parsing
        self.case_insensitive_names: Set[str] = set()
        self.field_alias_map: Dict[str, str] = {}
        self.attr_alias_map: Dict[str, str] = {}
        self.error_hooks: Dict[Type[Exception], Callable] = {}
        self.data_first_search = None
        self.addition_type = None
        self.name = get_obj_name(obj)
        self.is_local = is_local_var(obj)
        self.setup()

    def make_context(self, context=None, force_error: bool = False):
        return self.options.make_context(context=context, force_error=force_error)

    @property
    def kwargs(self):
        return {}

    def setup(self):
        self.generate_fields()
        self.generate_aliases()
        self.validate_fields()
        self.parse_addition_type()
        self.assign_search_strategy()

    def parse_addition_type(self):
        if self.options.addition and not isinstance(self.options.addition, bool):
            # we treat it as a type
            # including typing.GenericAlias / ForwardRef
            self.addition_type = self.parse_annotation(annotation=self.options.addition)

    def parse_annotation(self, annotation):
        return self.rule_cls.parse_annotation(
            annotation=annotation,
            forward_refs=self.forward_refs,
            global_vars=self.globals,
            force_clear_refs=self.is_local
        )

    @cached_property
    def property_fields(self):
        fields = {}
        for key, field in self.fields.items():
            if field.property:
                fields[key] = field
        return fields

    def validate_fields(self):
        pass
        # if self.options.allowed_runtime_options:
        #     if "__options__" in self.fields:
        #         raise ValueError(
        #             f"{self.obj} did not specify no_runtime_options=True, "
        #             f'so name "__options__" is used to passing runtime options, which is conflicted'
        #             f' with field {self.fields["__options__"]}'
        #         )

    def _get_field_from(self, fields: dict, key: str) -> Optional[ParserField]:
        if key in fields:
            return fields[key]
        if key in self.field_alias_map:
            return fields[self.field_alias_map[key]]
        if not key.islower() and key.lower() in self.case_insensitive_names:
            # avoid recursive
            return self._get_field_from(fields, key.lower())
        return None

    # def get_input_field(self, key: str) -> Optional[SchemaField]:
    #     return self._get_field_from(self.input_fields, key)

    def get_field(self, key: str) -> Optional[ParserField]:
        return self._get_field_from(self.fields, key)

    def get_attrs(self, data: Union[list, tuple, set, dict, str]):
        if isinstance(data, dict):
            return {self.get_attrs(key): val for key, val in data.items()}
        elif isinstance(data, (list, tuple, set)):
            return type(data)([self.get_attrs(v) for v in data])
        return self.attr_alias_map.get(data, data)

    def get_attname(self, key: str) -> Optional[str]:
        if key in self.attr_alias_map:
            return self.attr_alias_map[key]
        if not key.islower() and key.lower() in self.case_insensitive_names:
            # avoid recursive
            if key.lower() in self.attr_alias_map:
                return self.attr_alias_map[key.lower()]
        return None

    def assign_search_strategy(self):
        if self.options.data_first_search is not None:
            # explicitly specified
            self.data_first_search = self.options.data_first_search
            return
        if (
            self.case_insensitive_names
            or self.field_alias_map
            or self.options.ignore_required
            or self.options.addition
        ):
            # in those cases data first is faster than field first
            self.data_first_search = True
        else:
            self.data_first_search = False

    @property
    def rule_cls(self):
        return self.parser_field_cls.rule_cls

    @property
    def module_name(self):
        return self.obj.__module__

    @property
    def obj_name(self):
        return getattr(self.obj, "__qualname__", None) or getattr(self.obj, "__name__")

    @property
    def globals(self):
        if hasattr(self.obj, "__globals__"):
            # like a function
            return self.obj.__globals__
        return sys.modules[self.module_name].__dict__

    def __getitem__(self, item):
        return self.fields[item]

    def __contains__(self, item):
        return item in self.fields

    def resolve_forward_refs(self, local_vars=None, ignore_errors: bool = True):
        if not self.forward_refs:
            return False
        clear_refs = []
        resolved = False
        for name in list(self.forward_refs):
            ref, constraints = self.forward_refs[name]
            try:
                evaluate_forward_ref(ref, self.globals, local_vars)
                if ref.__forward_evaluated__:
                    # evaluated successfully, pop
                    value = ref.__forward_value__
                    if not isinstance(value, type):
                        # maybe some very foolish ForwardRef like
                        # a: "List[str, SomeClass]"
                        # we will just treat this nicely
                        __origin = get_origin(value)
                    else:
                        __origin = value
                    if __origin:
                        ref.__forward_value__ = self.rule_cls.parse_annotation(
                            annotation=value,
                            constraints=constraints,
                            global_vars=self.globals,
                            forward_refs=self.forward_refs,
                            forward_key=name,
                        )
                    else:
                        # maybe just ref to some const
                        # a: PositiveInt | '"some value"'
                        # we will not go through the ForwardRef
                        # as the typing star
                        ref.__forward_value__ = self.rule_cls.annotate(
                            type_=type(value),
                            constraints={"const": ref.__forward_value__},
                        )
                    resolved = True
                    if self.is_local:
                        clear_refs.append(ref)
                    self.forward_refs.pop(name)
            except Exception:
                if ignore_errors:
                    continue
                raise
        if resolved:
            for field in self.fields.values():
                field.resolve_forward_refs()
            # resolve for types
            self.addition_type, r = resolve_forward_type(self.addition_type)
        if self.is_local:
            # ForwardRef in local vars is not cachable
            # where typing is using a lru_cache
            # we should clear
            for ref in clear_refs:
                ref.__forward_evaluated__ = False
                ref.__forward_value__ = None
        return resolved

    @classmethod
    def validate_field_name(cls, name: str):
        return not name.startswith("_")

    def generate_fields(self):
        raise NotImplementedError

    def generate_aliases(self):
        alias_map = {}
        attr_alias_map = {}
        case_insensitive_names = set()

        for key, field in self.fields.items():
            if field.aliases:  # not contains the name
                for alias in field.aliases:
                    if key != alias:
                        if alias in alias_map:
                            raise exc.ConfigError(
                                f"{self.obj}: alias: [{repr(alias)}] "
                                f"conflict with field: [{repr(alias_map[alias])}]",
                                obj=self.obj,
                                field=field.name,
                            )
                        alias_map[alias] = key
                    # if field.attname != alias:
                    attr_alias_map[alias] = field.attname
                    # include equal

            if field.is_case_insensitive(self.options):
                case_insensitive_names.update(field.all_aliases)

        # for key, field in self.fields.items():
        #     if key in alias_map:
        #         raise ValueError(f'{self.obj}: alias: [{repr(key)}] conflict with field: [{repr(field)}]')

        if case_insensitive_names:
            for key, field in self.fields.items():
                if not field.is_case_insensitive(self.options):
                    lower_keys = set(a.lower() for a in field.aliases).union(
                        {key.lower()}
                    )
                    inter = case_insensitive_names.intersection(lower_keys)
                    if inter:
                        raise exc.ConfigError(
                            f"{self.obj}: case sensitive field: [{repr(key)}] "
                            f"conflict with case insensitive field in {inter}",
                            obj=self.obj,
                            field=field.name,
                        )

        # a: str = Field(alias_from=['a1', 'a2'], case_insensitive=True)
        # b: str = Field(alias='A2', alias_from=['A1'])

        self.field_alias_map = alias_map
        self.attr_alias_map = attr_alias_map
        self.case_insensitive_names = case_insensitive_names

        for key, field in self.fields.items():
            field.apply_fields(
                self.fields,
                # excluded_vars=self.exclude_vars,
                alias_map=alias_map,
            )

    @property
    def __ref__(self):
        return f"{self.obj.__module__}.{self.obj.__qualname__}"

    @property
    def cls(self):
        return self.obj if inspect.isclass(self.obj) else None

    def __call__(self, data: dict, context: RuntimeContext = None) -> dict:
        if not isinstance(data, dict):
            data = dict(data)
        self.resolve_forward_refs(ignore_errors=False)
        if not context:
            context = self.options.make_context(self.cls)
        result = self.parse_data(data, context=context)
        context.raise_error()
        # raise error if collected
        return result

    def parse_data(
        self,
        data: dict,
        context: RuntimeContext,
        as_attname: bool = None,
        excluded_keys: List[str] = None,
    ):
        options = context.options
        if options.max_params:
            if len(data) > options.max_params:
                context.handle_error(
                    exc.ParamsExceedError(
                        max_params=options.max_params, params_num=len(data)
                    )
                )
        if options.min_params:
            if len(data) < options.min_params:
                context.handle_error(
                    exc.ParamsLackError(
                        min_params=options.min_params, params_num=len(data)
                    )
                )
        dfs = (
            options.data_first_search
            if options.data_first_search is not None
            else self.data_first_search
        )
        if dfs:
            result = self.data_first_parse(
                data, context, excluded_keys=excluded_keys, as_attname=as_attname
            )
        else:
            result = self.field_first_parse(
                data, context, excluded_keys=excluded_keys, as_attname=as_attname
            )
        return result

    def parse_addition(self, key: str, value, context: RuntimeContext):
        if key in self.exclude_vars:
            # excluded vars cannot be carry in addition even if allowed
            return unprovided
        if context.options.addition is False:
            context.handle_error(exc.ExceedError(item=key, value=value))
            return unprovided
        if not context.options.addition:
            # None
            return unprovided
        # addition_type = options.addition if isinstance(options.addition, type) else self.addition_type
        addition_type = self.addition_type
        # we should just ignore the runtime addition type
        if not addition_type:
            return value

        options = context.options
        with context.enter(key) as new_context:
            try:
                value = new_context.transformer(value, addition_type)
            except Exception as e:
                error = exc.ParseError(
                    item=key, value=value, type=addition_type, origin_exc=e
                )
                if options.invalid_values == options.EXCLUDE:
                    context.collect_waring(error.formatted_message)
                    return unprovided
                elif options.invalid_values == options.PRESERVE:
                    context.collect_waring(error.formatted_message)
                else:
                    context.handle_error(error)
        return value

    def data_first_parse(
        self,
        data: dict,
        context: RuntimeContext,
        as_attname: bool = False,
        excluded_keys: List[str] = None,
    ):
        addition = {}
        result = {}
        dependencies = set()
        unprovided_fields = set()
        options = context.options

        for key, value in data.items():
            key = str(key)
            field = self.get_field(key)
            if not field:
                add_value = self.parse_addition(key, value, context=context)
                if not unprovided(add_value):
                    addition[key] = add_value
                continue

            name = field.attname if as_attname else field.name

            if field.is_no_input(value, options=options):
                # no input field does not take input from __init__
                # but can still apply default
                default = field.get_default(options, defer=False)
                if not unprovided(default):
                    result[name] = default
                continue

            if (name in result) or (excluded_keys and name in excluded_keys):
                if options.ignore_alias_conflicts:
                    continue
                context.handle_error(exc.AliasConflictError(item=name, value=value))
                continue

            parsed = field.parse_value(value, context=context)
            if unprovided(parsed):
                continue

            result[name] = parsed

            if field.dependencies:
                dependencies.update(
                    field.attr_dependencies if as_attname else field.dependencies
                )

        if not options.ignore_required:
            # if required field is ignored. we do not need to check for required fields
            for key, field in self.fields.items():
                name = field.attname if as_attname else field.name
                if name in result:
                    continue
                if excluded_keys and name in excluded_keys:
                    continue
                unprovided_fields.add(name)
                if field.is_required(options=options):
                    context.handle_error(exc.AbsenceError(item=name))
                    continue
                default = field.get_default(options, defer=False)
                if not unprovided(default):
                    result[name] = default

        if dependencies:
            dependant = set(result)
            if excluded_keys:
                dependant.update(excluded_keys)

            diff = dependencies.difference(dependant)
            lack = dependencies.intersection(unprovided_fields)
            lack.update(diff)
            if lack:
                # some dependencies not provided
                context.handle_error(
                    exc.DependenciesAbsenceError(absence_dependencies=lack)
                )

        # check dependencies before addition

        if addition:
            result.update(addition)

        return result

    def field_first_parse(
        self,
        data: dict,
        context: RuntimeContext,
        as_attname: bool = False,
        excluded_keys: List[str] = None,
    ):
        if self.case_insensitive_names:
            _data = {}
            for k, v in data.items():
                k = str(k)
                if k.lower() in self.case_insensitive_names:
                    _data[k.lower()] = v
                else:
                    _data[k] = v
            data = _data

        result = {}
        used_alias = set()
        dependencies = set()
        unprovided_fields = set()
        options = context.options

        for key, field in self.fields.items():
            value = unprovided
            name = field.attname if as_attname else field.name

            if excluded_keys and name in excluded_keys:
                continue

            if options.ignore_alias_conflicts:
                for alias in field.all_aliases:
                    if alias in data:
                        value = data[alias]
                        break
            else:
                for alias in field.all_aliases:
                    if alias in data:
                        if unprovided(value):
                            value = data[alias]
                        else:
                            context.handle_error(exc.AliasConflictError(item=name))
                            break

            if unprovided(value):
                unprovided_fields.add(name)
                if field.is_required(options=options):
                    context.handle_error(exc.AbsenceError(item=name))
                    continue
                default = field.get_default(options, defer=False)
                # we don't catch this error for now
                # because default function is "server" function
                # if the default goes wrong, it should directly raise to the user
                if not unprovided(default):
                    result[name] = default
                continue

            used_alias.update(field.all_aliases)
            # even if field is no-input, it can still set default (by developer, no by input)
            if field.is_no_input(value, options=options):
                # no input field does not take input from __init__
                # but can still apply default
                default = field.get_default(options, defer=False)
                if not unprovided(default):
                    result[name] = default
                continue

            parsed = field.parse_value(value, context=context)
            if unprovided(parsed):
                continue

            result[name] = parsed
            if field.dependencies:
                dependencies.update(
                    field.attr_dependencies if as_attname else field.dependencies
                )

        if dependencies:
            dependant = set(result)
            if excluded_keys:
                dependant.update(excluded_keys)

            diff = dependencies.difference(dependant)
            lack = dependencies.intersection(unprovided_fields)
            lack.update(diff)
            if lack:
                # some dependencies not provided
                context.handle_error(
                    exc.DependenciesAbsenceError(absence_dependencies=lack)
                )

        # check dependencies before addition

        if options.addition is not None:
            # that we cannot ignore addition here
            addition = {}
            for k, v in data.items():
                if k in used_alias:
                    continue
                add_value = self.parse_addition(k, v, context=context)
                if not unprovided(add_value):
                    addition[k] = add_value
            result.update(addition)

        return result



from decimal import Decimal
from enum import Enum
from datetime import datetime, timedelta, date
from uuid import UUID
from typing import Union, Optional, Tuple, List, Set, Mapping, \
    Dict, Type, Callable, Any, TYPE_CHECKING, Iterator, ClassVar
from .utils.compat import Literal, Annotated, Final, ForwardRef
from .parser.rule import Lax, Rule

# from typing import TypeVar
# T = TypeVar('T')


class Number(Rule):
    primitive = "number"

    @classmethod
    def check_type(cls, t):
        # must be an iterable
        assert issubclass(t, (int, float, Decimal))


class Array(Rule):
    __origin__ = list  # use list instead of abc.Iterable
    primitive = "array"

    # def __class_getitem__(cls, item) -> Type["Array"]:
    #     if not isinstance(item, tuple):
    #         item = (item,)
    #     return cls.annotate(cls.__origin__, *item)

    @classmethod
    def check_type(cls, t):
        # must be an iterable
        assert hasattr(t, "__iter__")


class Object(Rule):
    __origin__ = dict
    primitive = "object"

    def __class_getitem__(cls, params):
        if len(params) != 2:
            raise TypeError(
                f"{cls} should use {cls}[KeyType, ValType] with 2 params, got {params}"
            )
        return cls.annotate(cls.__origin__, *params)

    @classmethod
    def check_type(cls, t):
        # must be an iterable
        assert hasattr(t, "__iter__") and hasattr(t, "items")


class Float(float, Number):
    pass


class Int(int, Number):
    pass


class Str(str, Rule):
    pass


class Bool(Rule):
    __origin__ = bool


class Null(Rule):
    __origin__ = type(None)
    primitive = "null"
    # const = None


class PositiveInt(Int):
    gt = 0


class NaturalInt(Int):
    ge = 0


class PositiveFloat(Float):
    gt = 0


class NegativeFloat(Float):
    lt = 0


class NanFloat(Float):
    @classmethod
    def post_validate(cls, value, options=None):
        import math

        if not math.isnan(value):
            # do not use const = float('nan')
            # cause NaN can not use equal operator
            from .utils import exceptions as exc
            raise exc.ConstraintError(constraint="const", constraint_value=float("nan"))
        return value


class InfinityFloat(Float):
    enum = [float("inf"), float("-inf")]


AbnormalFloat = NanFloat ^ InfinityFloat

NormalFloat = Float & ~AbnormalFloat


class Zero(Rule):
    const = 0


Divisor = Float & ~Zero


class NegativeInt(Int):
    lt = 0


class Timestamp(Float):
    """
    timestamps to represent datetime and date
    """

    ge = 0
    format = "timestamp"

    @classmethod
    def pre_validate(cls, value, options=None):
        import datetime

        if isinstance(value, datetime.datetime):
            value = value.timestamp()
        elif isinstance(value, datetime.timedelta):
            value = value.total_seconds()
        return value


def round_number(precision: int = 0, num_type: type = float):
    class RoundNumber(num_type, Rule):
        decimal_places = Lax(precision)

    return RoundNumber


def enum_array(
    item_enum: Union[Type[Enum], list, tuple, set],
    item_type=None,
    array_type=list,
    unique: bool = False,
) -> Type[Array]:
    """
    Make an array type, which item is one of the enum value
    """
    if isinstance(item_enum, Enum):
        EnumItem = item_enum
    else:

        class EnumItem(Rule):
            __origin__ = item_type
            enum = item_enum

    class EnumArray(Array):
        __origin__ = array_type
        __args__ = (EnumItem,)
        __ellipsis_args__ = issubclass(array_type, tuple)
        unique_items = unique

    return EnumArray


class SlugStr(Str):
    """
    Slug str or URI
    """

    format = "slug"
    regex = r"[a-z0-9]+(?:-[a-z0-9]+)*"


# class Secret(Str):
#     """
#     Slug str or URI
#     """
#     def __repr__(self):
#         return f'{self.__class__.__name__}("%s")' % ('*' * 6)
#
#     def __str__(self):
#         return f'{self.__class__.__name__}("%s")' % ('*' * 6)


class Year(Int):
    ge = 1
    le = 9999


class Month(Int):
    ge = 1
    le = 12


class Day(Int):
    ge = 1
    le = 31


class Week(Int):
    ge = 1
    le = 53


class WeekDay(Int):
    ge = 1
    le = 7


class Quarter(Int):
    ge = 1
    le = 4


class Hour(Int):
    ge = 0
    le = 23


class Minute(Int):
    ge = 0
    le = 59


class Second(Int):
    ge = 0
    le = 59


class Datetime(datetime, Rule):
    primitive = "string"
    format = "datetime"


class Date(date, Rule):
    primitive = "string"
    format = "date"


class Timedelta(timedelta, Rule):
    primitive = "string"
    format = "duration"


class EmailStr(Str):
    format = "email"
    regex = r"([A-Za-z0-9]+[.-_])*[A-Za-z0-9]+@[A-Za-z0-9-]+(\.[A-Z|a-z]{2,})+"


# from pathlib import Path
#
#
# class FilePath(Path, Rule):
#     format = 'file-path'
#     is_dir: bool = False
#     is_abs: bool = None
#     is_link: bool = None
#     is_exists: bool = None
#     max_size: int
#     min_size: int
#
#
# class Directory(Path, Rule):
#     format = 'directory'
#     is_dir: bool = True
#     is_abs: bool = None
#     is_link: bool = None
#     is_exists: bool = None
#     max_files: int
#     min_files: int
